"""
This module provides the main `Simulation` class to perform the growth KMC.
"""


from dataclasses import dataclass, field
from typing import Optional, Tuple, IO
from functools import cached_property
from io import StringIO
import logging
import warnings

import numpy as np
import scipy # type: ignore

from .containers import Solvent, Growth, CubicLattice, KthNearest

logger = logging.getLogger(__name__)


@dataclass
class Simulation:
    """
    Simulation class.
    Main method is `Simulation.perform`, which performs a crystal growth simulation using Kinetic Monte Carlo

    Attributes:
        lattice (CubicLattice):
            The lattice on which to perform KMC
        interactions (KthNearest):
            The interactions between lattice sites
        solvent (Solvent):
            The solvent surrounding the initial spherical crystal
        growth (Growth):
            Growth parameters for the simulation
        seed (int):
            Optional seed to provide to the simulation for random number generation. Defaults to 0
        generator (np.random.Generator):
            Generator for the simulation. This attribute is set after initialization
        lattice_points (np.ndarray):
            Lattice points in the supercell. This attribute is set after initialization
        bounds (np.ndarray):
            Supercell bounds. This attribute is set after initialization
        surface_density (float):
            Surface density of the initial spherical seed in units of molecules per area. This attribute is set after
            initialization
    """

    lattice: CubicLattice
    """lattice inside supercell"""

    interactions: KthNearest
    """interactions between lattice sites"""

    solvent: Solvent
    """solvent surrounding initial seed"""

    growth: Growth
    """Growth parameters"""

    seed: Optional[int] = 0
    """Optional seed for RNG"""

    generator: np.random.Generator = field(init=False)
    """Random number generator"""

    lattice_points: np.typing.NDArray[np.floating] = field(init=False)
    """Lattice points inside supercell"""

    bounds: np.typing.NDArray[np.floating] = field(init=False)
    """Super cell bounds"""

    surface_density: float = field(init=False)
    """Initial spherical surface density"""

    def __post_init__(self):

        self.generator = np.random.default_rng(seed=self.seed)
        self.lattice_points, self.bounds = self.lattice.initialize_simulation()

    @cached_property
    def hamiltonian(self) -> scipy.sparse.csr_matrix:

        r"""
        Defines the Hamiltonian matrix $\mathbf{Q}$, which defines the energy function:
        $$E(x) = \frac{1}{2}\mathbf{x}^\intercal\mathbf{Q}\mathbf{x}$$

        Returns:
            scipy.sparse.csr_matrix: Interaction matrix $\mathbf{Q}$
        """

        return self.interactions.get_hamiltonian(self.lattice_points, self.bounds)

    @cached_property
    def adjacency_matrix(self) -> scipy.sparse.csr_matrix:

        r"""
        Defines the adjacency matrix $\mathbf{A}$, where $A_{ij} = 1$ if sites $i$ and $j$ have an interaction.

        Returns:
            scipy.sparse.csr_matrix: Adjacency matrix $\mathbf{A}$
        """

        return scipy.sparse.csr_matrix(
            (np.ones_like(self.hamiltonian.data), self.hamiltonian.indices, self.hamiltonian.indptr),
            shape=self.hamiltonian.shape
        )

    @cached_property
    def num_neighbors(self) -> int:

        r"""
        Coordination number of the lattice.

        Returns:
            int: Coordination number, counting all neighbors
        """

        num_neighbors_per_site = self.adjacency_matrix.sum(axis=0)
        if not np.isclose(num_neighbors_per_site.std(), 0.0):
            warnings.warn(
                "Non-periodic lattice. There is likely problems with periodic boundary conditions. This is probably "
                "fine if your crystal does not grow to the boundary"
            )
        return num_neighbors_per_site.mean()

    def get_lammps_dump_str(
        self,
        types: np.typing.NDArray[np.floating],
        step: int,
        t: float,
        fmt: Tuple[str, str, str, str, str] = ("%.0f", "%.0f", "%.4f", "%.4f", "%.4f")
    ) -> str:

        r"""
        Shortcut method for getting a LAMMPS-style dump str, which looks like:

        ```
        ITEM: TIMESTEP
        0 0.0
        ITEM: NUMBER OF ATOMS
        1874
        ITEM: BOX BOUNDS xy xz xx yy zz
        0.0 318.08 0.0
        0.0 318.08 0.0
        0.0 303.165 0.0
        ITEM: ATOMS id type x y z
        39105 1 113.6000 131.7760 151.5825
        39189 1 113.6000 140.8640 131.3715
        39191 1 113.6000 140.8640 138.1085
        39193 1 113.6000 140.8640 144.8455
        39195 1 113.6000 140.8640 151.5825
        39197 1 113.6000 140.8640 158.3195
        ...
        ```

        at each timestep

        Arguments:
            types (np.ndarray):
                State matrix $\mathbf{x}$, where $x_i = 1$ if site $i$ is occupied and $0$ else.
            step (int):
                Current step in the simulation.
            t (float):
                Current time in the simulation.
            fmt (Tuple[str, str, str, str, str]):
                Sequence of format specifiers for LAMMPS-style dumping.

        Returns:
            str: The LAMMPS-style dump string
        """

        mask = types == 1

        ids = np.arange(len(self.lattice_points))
        occupied_ids = ids[mask].reshape(-1, 1)
        occupied_sites = self.lattice_points[mask, :]

        header = "\n".join([
            "ITEM: TIMESTEP",
            f"{step:.0f} {t}",
            "ITEM: NUMBER OF ATOMS",
            f"{len(occupied_ids):.0f}",
            "ITEM: BOX BOUNDS xy xz xx yy zz",
            f"0.0 {self.bounds[0]} 0.0",
            f"0.0 {self.bounds[1]} 0.0",
            f"0.0 {self.bounds[2]} 0.0",
            "ITEM: ATOMS id type x y z"
        ])

        data = np.concatenate((occupied_ids, np.ones_like(occupied_ids), occupied_sites), axis=1)

        with StringIO() as string_io:
            np.savetxt(string_io, data, fmt=fmt, comments="", header=header) # type: ignore
            return string_io.getvalue()

    @property
    def kappa(self) -> float:

        r"""
        Shorthand when computing dynamic evaporation prefactor, i.e.:

        $$\nu_t = \kappa / \left\langle \exp\left(\beta\Delta E_\text{evap}\right)\right\rangle$$

        Returns:
            float: $\kappa$
        """

        # only defined if the surface density has been calculated or not
        if not hasattr(self, "surface_density"):
            raise ValueError("surface density not yet calculated")

        return (4.0 / 3.0 * np.pi * self.lattice.density) ** (1 / 3) * self.solvent.diffusivity * \
            self.solvent.solubility_limit / (self.surface_density * self.growth.desired_size ** (1 / 3))

    def get_interface(
        self,
        types: np.typing.NDArray[np.floating]
    ) -> Tuple[np.typing.NDArray[np.integer], np.typing.NDArray[np.integer]]:

        r"""
        Function for computing interfacial solvent and solid sites.
        $\mathbf{A}\mathbf{x}$ counts the number of currently occupied neighbors.

        Arguments:
            types (np.ndarray): State vector $\mathbf{x}$

        Returns:
            Tuple[np.ndarray, np.ndarray]: IDs of solid sites and IDs of solvent sites at interface
        """

        occupied_neighbor_count = self.adjacency_matrix @ types
        solid_sites, = np.where(types * (self.num_neighbors - occupied_neighbor_count) > 0)
        solvent_sites, = np.where((1.0 - types) * occupied_neighbor_count > 0)

        return solid_sites, solvent_sites

    def perform(self, dump_file: IO, dump_every: int):

        """
        Main method for performing a Kinetic Monte Carlo simulation. This is performable by calling something like:

        ```py
        simulation = Simulation(...)
        with open("kmc.dump", "w") as file:
            simulation.perform(dump_file=file, dump_every=100)
        ```

        or any other IO-like object, such as StringIO or BytesIO

        Arguments:

            dump_file (IO):
                `IO` instance where user will store dump information
            dump_every (int):
                How often to dump information
        """

        # initialize a spherical crystal with specified radius
        center = self.lattice_points.mean(axis=0)
        types = (np.linalg.norm(self.lattice_points - center, axis=1) <= self.growth.initial_radius).astype(float)

        solid_sites, _ = self.get_interface(types)
        self.surface_density = len(solid_sites) / self.growth.initial_surface_area

        t = 0.0
        total_energy = 0.5 * types.T @ self.hamiltonian @ types
        for step in range(self.growth.num_steps):

            # if types.mean() is 0, entire cell is unoccupied, so no simulation
            # similarly, if types.mean() is 1, entire cell is occupied
            occupancy = types.mean()
            if not 0.0 < occupancy < 1.0:
                logging.error("simulation killed, no interface detected")
                return

            # only want dumps at a small frequency
            if not step % dump_every:
                positions = self.get_lammps_dump_str(types, step, t)
                print(positions, file=dump_file, end="")
                logging.info("simulation info", extra={
                    "step": step, "t": t, "total_energy": total_energy, "occupancy": occupancy
                })
                dump_file.flush()

            solid_sites, solvent_sites = self.get_interface(types)

            # need to rescale evaporation rates after calculating
            # this is the dynamically updated prefactor
            evaporation_barriers = -self.hamiltonian[solid_sites, :] @ types
            evaporation_rates = np.exp(-self.solvent.beta * evaporation_barriers)
            rate_prefactor = self.kappa / evaporation_rates.mean()
            evaporation_rates = rate_prefactor * evaporation_rates

            radius = (0.75 * self.lattice.molecular_volume * types.sum() / np.pi) ** (1 / 3)
            adsorption_rates = np.ones_like(solvent_sites) * self.solvent.diffusivity * \
                self.solvent.solubility_limit / (self.surface_density * radius)

            # concatenate events, so we can pick event according to residence time algorithm
            events = np.concatenate((solid_sites, solvent_sites))
            rates = np.concatenate((evaporation_rates, adsorption_rates))
            total_rate = rates.sum()

            # pick event and advance time according to residence time algorithm
            event = self.generator.choice(events, p=rates / total_rate)

            # if there was an evaporation event, we've already computed \Delta E_evap
            if types[event]:
                change_in_energy_evap = evaporation_barriers[np.where(events == event)].item()
                total_energy += change_in_energy_evap
            # if it was an adsorption event, need to compute it
            elif not types[event]:
                change_in_energy_evap = -(self.hamiltonian[event, :] @ types).item()
                total_energy += -change_in_energy_evap
            else:
                raise ValueError

            types[event] = 1.0 - types[event]
            t += self.generator.exponential(scale=1.0 / total_rate)



"""
This module provides some extraneous utilities
"""


from enum import Enum, auto
from hashlib import sha1
import warnings

import numpy as np


class Units(Enum):
    """
    Available units. See https://docs.lammps.org/units.html for details
    """

    real = auto()
    metal = auto()
    si = auto()
    cgs = auto()
    electron = auto()
    micro = auto()
    nano = auto()

    def boltzmann_constant(self):

        r"""
        $k_B$ in whatever chosen unit system.

        Returns:
            float: [Boltzmann constant](https://en.wikipedia.org/wiki/Boltzmann_constant)
        """

        if self == Units.real:
            return 1.987e-3

        if self == Units.metal:
            return 8.617e-5

        if self == Units.si:
            return 1.381e-23

        if self == Units.cgs:
            return 1.381e-16

        if self == Units.electron:
            return 3.167e-6

        if self == Units.micro:
            return 1.381e-6

        if self == Units.nano:
            return 1.381e-2

        raise ValueError("invalid unit system")


def temp_to_beta(temperature: float, units: Units):
    r"""
    helper class to convert temperature to thermodynamic $\beta$ for a given choice of units

    Arguments:
        temperature (float):
            Absolute temperature $T$
        units (Units):
            Chosen unit system

    Returns:
        float: Thermodynamic $\beta = 1/(k_BT)$
    """

    if temperature == 0:
        warnings.warn("You have a zero temperature!")
        return np.inf

    if temperature < 0:
        warnings.warn("You have a negative temperature!")

    return 1.0 / (units.boltzmann_constant() * temperature)


def array_to_hex(x: np.typing.NDArray[np.floating]) -> str:

    """
    Method for computing an array into a unique hex string using the
    [SHA-1 function](https://en.wikipedia.org/wiki/SHA-1)

    Arguments:
        x (np.ndarray): Desired array to hex-ify

    Returns:
        str: Hex-ified string
    """

    return sha1(x.tobytes()).hexdigest()



r"""
.. include:: ../README.md

# Examples

## 📦 Simple cubic growth

Below is an example of the growth of a small simple cubic lattice, and outputting the simulation info to stdout

```py
.. include:: ../examples/simple_cubic.py
```

This prints the simulation dump in LAMMPS-style dump format:

```
ITEM: TIMESTEP
0 0.0
ITEM: NUMBER OF ATOMS
33
ITEM: BOX BOUNDS xy xz xx yy zz
0.0 5.0 0.0
0.0 5.0 0.0
0.0 5.0 0.0
ITEM: ATOMS id type x y z
12 1 0.0000 2.0000 2.0000
31 1 1.0000 1.0000 1.0000
32 1 1.0000 1.0000 2.0000
...
ITEM: TIMESTEP
100 2.39212295038683e-07
ITEM: NUMBER OF ATOMS
95
ITEM: BOX BOUNDS xy xz xx yy zz
0.0 5.0 0.0
0.0 5.0 0.0
0.0 5.0 0.0
ITEM: ATOMS id type x y z
0 1 0.0000 0.0000 0.0000
1 1 0.0000 0.0000 1.0000
2 1 0.0000 0.0000 2.0000
...
```

and so on. This can then be redirected to a file, and visualized by many trajectory viewing softwares, including
[Open Visualization Tool](https://www.ovito.org/) (OVITO),
[Visual Molecular Dynamics](https://www.ks.uiuc.edu/Research/vmd/) (VMD), and
[more](https://en.wikipedia.org/wiki/List_of_molecular_graphics_systems).

## ⚛ Visualizing with OVITO

Below is a simple example of growing a simple cubic crystal starting from a spherical seed, dumping to a file
`cube.dump`, and compressing the dump file to `cube.dump.gz`.

```py
.. include:: ../examples/ovito_visualization.py
```

This dump file (and its compressed version) can be loaded into OVITO for visualization! Simply:

- Open OVITO
- Load the dump file (`File` ➡ `Load File` ➡ `cube.dump.gz` ➡ `Open`)

Your window should look like this:

<p align="center">
<img src="https://raw.githubusercontent.com/jwjeffr/cgkmc/refs/heads/main/examples/ovito_steps/step1.jpg" alt="Loading in dump file" style="width: 50%;">
</p>

The particles here are not particularly useful for visualization. Instead, we can create a surface mesh. Click:

- `Add modification` (top right, under the dump file name)
- `Construct surface mesh` (Scroll down to `Visualization` category)

Your window should now look like this:

<p align="center">
<img src="https://raw.githubusercontent.com/jwjeffr/cgkmc/refs/heads/main/examples/ovito_steps/step2.jpg" alt="Constructing surface mesh" style="width: 50%;">
</p>

The particles here are still a bit of an eyesore. You can get rid of them by clicking `Particles` under the
`Visual elements` tab, also under the dump file name. After doing this, your window should look like this:

<p align="center">
<img src="https://raw.githubusercontent.com/jwjeffr/cgkmc/refs/heads/main/examples/ovito_steps/step3.jpg" alt="Hiding particles" style="width: 50%;">
</p>

and now this looks much better! You can now press the play button (bottom center) to view the animation. Your window at
the final frame should look like this:

<p align="center">
<img src="https://raw.githubusercontent.com/jwjeffr/cgkmc/refs/heads/main/examples/ovito_steps/step4.jpg" alt="Final frame" style="width: 50%;">
</p>

This final frame makes sense! The $\\{100\\}$ surfaces have the lowest surface energy in simple cubic crystals with
first nearest neighbor interactions, so the crystal should grow into something roughly rectangular.

## 📝 Using a logger

Below is an example of using a logger to grab information from the simulation. `cgkmc` uses Python's native `logging`
library (with docs [here](https://docs.python.org/3/library/logging.html)), so one can use a `logging` config to grab
simulation information.

```py
.. include:: ../examples/writing_log.py
```

Here, `occupancy` is the proportion of occupied sites. This creates a log file that looks like:

```
2025-03-30 12:36:26,790 INFO:simulation info TIME=6.758222371127191e-06 ENERGY=-3518.0 OCCUPANCY=0.491
2025-03-30 12:36:26,817 INFO:simulation info TIME=6.787981905980638e-06 ENERGY=-3547.0 OCCUPANCY=0.494
2025-03-30 12:36:26,843 INFO:simulation info TIME=6.813314575607091e-06 ENERGY=-3559.0 OCCUPANCY=0.496
```

and so on. This can be parsed using Python's built-in
[regular expression](https://en.wikipedia.org/wiki/Regular_expression) parser, and the results can be visualized:

```py
.. include:: ../examples/parsing_log.py
```

which generates the plot below:

<p align="center">
<img src="https://raw.githubusercontent.com/jwjeffr/cgkmc/refs/heads/main/examples/energy.png" alt="Energy vs. time">
</p>

You can get very creative with this! My preferred is logging in [JSON Lines format](https://jsonlines.org/).
mCoding has a very great YouTube video on this, which is [here](https://www.youtube.com/watch?v=9L77QExPmI0).

## 🧪 Case study: PETN

### What is PETN?

[Pentaerythritol tetranitrate](https://en.wikipedia.org/wiki/Pentaerythritol_tetranitrate) (PETN) is a very well-studied
organic crystalline system. Here, we will run `cgkmc` on a PETN system to predict the morphology of a PETN nanocrystal.

### PETN Parameters

Following work by Singh. et al. [here](https://doi.org/10.1021/acs.cgd.3c01487), we can reduce PETN molecules to
individual effective atoms inside a body-tetragonal unit cell, with lattice parameters $a = b = 9.087\;\text{Å}$,
$c = 6.738\;\text{Å}$, and "atomic" basis $\\{(0, 0, 0), (1/2, 1/2, 1/2)\\}$. Atomic is in quotes here because
the particles are molecules, not atoms. Additionally, we can use the interaction energies `(-0.294, -0.184, -0.002)` and
corresponding cutoff distances `(7.0, 7.5, 9.5)`, respectively in $\text{eV}$ and $\text{Å}$.

### Script

A script for the simulation of such a system is below:

```py
from pathlib import Path

from cgkmc import simulations, containers, utils


def main():

    simulation = simulations.Simulation(
        lattice=containers.CubicLattice(
            dimensions=(62, 62, 140),
            lattice_parameters=(9.087, 9.087, 6.738),
            atomic_basis=[
                [0.0, 0.0, 0.0],
                [0.5, 0.5, 0.5]
            ]
        ),
        interactions=simulations.KthNearest(
            cutoffs=(7.0, 7.5, 9.5),
            interaction_energies=(-0.294, -0.184, -0.002),
            use_cache=True
        ),
        solvent=containers.Solvent(
            beta=utils.temp_to_beta(temperature=300, units=utils.Units.metal),
            diffusivity=1.0e+11,
            solubility_limit=1.0e-4
        ),
        growth=containers.Growth(
            initial_radius=75.0,
            num_steps=1_000_000,
            desired_size=40_000
        )
    )

    with Path("petn.dump").open("w") as file:
        simulation.perform(file, dump_every=1_000)


if __name__ == "__main__":

    main()
```

This generates a dump file that we can analyze in OVITO! The animation is below:

<div style="padding:75% 0 0 0;position:relative;">
<iframe
    src="https://player.vimeo.com/video/1078144061?h=446c9e7099&amp;badge=0&amp;autopause=0&amp;player_id=0&amp;app_id=58479" frameborder="0" allow="autoplay; fullscreen; picture-in-picture; clipboard-write; encrypted-media" style="position:absolute;top:0;left:0;width:100%;height:100%;"
    title="petn"
>
</iframe>
</div>
<script src="https://player.vimeo.com/api/player.js">
</script>

Some clear crystal morphology jumps out, namely $\\{110\\}$ surfaces. See a more full study
[here](https://arxiv.org/abs/2509.25490)!

## ⌨ Using the command line interface

We also provide an option to use a command-line interface (CLI)! For example, for the simple cubic example from earlier,
the input file looks like:

```json
.. include:: ../examples/simple_cubic.json
```

and the resulting CLI call looks like:

```bash
cgkmc --input_file input_file.json --dump_file dump_file.dump --dump_every 1000
```

which runs a simulation, and dumps molecular coordinates every 1000 steps. You can also specify an optional log file
with `--log_file`, which gives you access to dynamical variables like occupancy and energy.

This CLI automatically comes with `cgkmc`. No additional building is necessary! You can also run:

```bash
cgkmc --help
```

for more info!
"""

__version__ = "0.1.1"
__authors__ = ["Jacob Jeffries"]
__author_emails__ = ["jwjeffr@clemson.edu"]
__url__ = "https://github.com/jwjeffr/cgkmc"

from . import containers as containers
from . import simulations as simulations
from . import utils as utils

import argparse
from pathlib import Path
import json
import logging


def cli():

    """
    @private
    """

    parser = argparse.ArgumentParser(
        prog="cgkmc",
        description="Command line interface for Crystal Growth Kinetic Monte Carlo (cgkmc)",
        epilog="See https://jwjeffr.github.io/cgkmc/ for full documentation!"
    )
    parser.add_argument("-i", "--input_file", type=Path, help="path to input json file")
    parser.add_argument("-d", "--dump_file", type=Path, help="file to dump coordinates to"),
    parser.add_argument("-n", "--dump_every", type=int, help="how often to dump coordinates")
    parser.add_argument("-l", "--log_file", type=Path, help="optional log file", required=False, default=None)
    args = parser.parse_args()

    if args.log_file:
        with args.log_file.open("w"):
            pass
        logging.basicConfig(
            filename=args.log_file,
            level=logging.INFO,
            format='%(asctime)s %(levelname)s:%(message)s TIME=%(t)s ENERGY=%(total_energy)s OCCUPANCY=%(occupancy)s'
        )

    with args.input_file.open("r") as file:
        input_dict = json.load(file)

    simulation = simulations.Simulation(
        lattice=containers.CubicLattice(**input_dict["lattice"]),
        interactions=containers.KthNearest(**input_dict["interactions"]),
        solvent=containers.Solvent(**input_dict["solvent"]),
        growth=containers.Growth(**input_dict["growth"])
    )

    with args.dump_file.open("w") as file:
        simulation.perform(dump_file=file, dump_every=args.dump_every)



"""
This module has container classes to propagate data through the KMC simulation. These container classes define the
solvent, growth, and lattice of the simulation, as well as the interaction energies between neighbors.
"""


from dataclasses import dataclass
from typing import Tuple, Optional
import logging
from pathlib import Path

import numpy as np
import scipy # type: ignore

from .utils import array_to_hex

logger = logging.getLogger(__name__)


@dataclass
class Solvent:
    r"""
    Solvent container class.

    Temperature (or $\beta$), diffusivity, and solubility limit fully define the solvent. $\beta$ should match the
    units of the interaction energies you specify. See `utils.Units` class for available energy units, and use
    `utils.temp_to_beta` to get $\beta$ from temperature in Kelvin.

    Attributes:
        beta (float):
            Thermodynamic $\beta$ defined as $1/(k_BT)$, where $k_B$ is the
            [Boltzmann constant](https://en.wikipedia.org/wiki/Boltzmann_constant) and $T$ is temperature
        diffusivity (float):
            Diffusivity of solute in liquid phase $D$
        solubility_limit (float):
            Solubility limit, or equivalently concentration of solute in liquid phase $n_\infty$
    """

    beta: float
    r"""thermodynamic $\beta$"""

    diffusivity: float
    r"""diffusivity of solute $D$"""

    solubility_limit: float
    r"""solubility limit $n_\infty$"""


@dataclass
class Growth:
    r"""
    Growth container class.
    Mimics experimental controls, i.e. initial crystal size, amount of time we grow, and the final size we want.

    Attributes:
        initial_radius (float):
            Initial radius of spherical seed in physical units, or the same units as the specified lattice parameters
        num_steps (int):
            Number of KMC steps to perform
        desired_size (int):
            The desired number of molecules in the final crystal. The final crystal will not have this exact number of
            molecules, but increasing this parameter will increase the final size.
    """

    initial_radius: float
    r"""initial radius $R_0$"""

    num_steps: int
    r"""number of KMC steps $N_\text{steps}$"""

    desired_size: int
    r"""desired size of final crystal $N_*$"""

    def __post_init__(self):

        if self.initial_radius <= 0:
            raise ValueError(f"{self.__class__.__name__}.initial_radius should be positive")

        if self.num_steps <= 0 or not isinstance(self.num_steps, int):
            raise ValueError(f"{self.__class__.__name__}.num_steps should be a positive integer")

        if self.desired_size <= 0 or not isinstance(self.desired_size, int):
            raise ValueError(f"{self.__class__.__name__}.desired_size should be a positive integer")

    @property
    def initial_surface_area(self) -> float:
        r"""
        Shortcut property for computing surface area.
        Crystal is assumed to be spherical, so surface area = $4\pi\times (\text{radius})^2$
        """

        return 4.0 * np.pi * self.initial_radius ** 2


@dataclass
class CubicLattice:

    r"""
    Lattice container class, which creates the resulting cubic lattice.

    Attributes:
        dimensions (np.ndarray):
            Dimensions of the cubic lattice in terms of unit cells. For a $10\times 15\times 20$ lattice, for example,
            specify `dimensions=np.array([10, 15, 20])`.
        lattice_parameters (np.ndarray):
            Lattice parameters of the lattice $a$, $b$, and $c$.
        atomic_basis (np.ndarray):
            Atomic basis of the solid phase's unit cell.
            Should be an array with size $(\text{number of molecules per unit cell}, 3)$
    """

    dimensions: np.typing.NDArray[np.integer]
    r"""Dimensions of the cubic lattice in terms of unit cells"""

    lattice_parameters: np.typing.NDArray[np.floating]
    r"""Lattice parameters of the lattice $a$, $b$, and $c$."""

    atomic_basis: np.typing.NDArray[np.floating]
    r"""Atomic basis of the solid phase's unit cell."""

    def __post_init__(self):
        # turn objects into arrays if they're not already arrays
        if not isinstance(self.dimensions, np.ndarray):
            self.dimensions = np.array(self.dimensions, dtype=int)
        if not isinstance(self.lattice_parameters, np.ndarray):
            self.lattice_parameters = np.array(self.lattice_parameters, dtype=float)
        if not isinstance(self.atomic_basis, np.ndarray):
            self.atomic_basis = np.array(self.atomic_basis, dtype=float)

        if len(self.dimensions.shape) != 1:
            raise ValueError(
                f"Invalid lattice geometry. {self.__class__.__name__}.dimensions should have shape "
                f"(number of spatial dimensions,)"
            )

        if len(self.lattice_parameters.shape) != 1:
            raise ValueError(
                f"Invalid lattice geometry. {self.__class__.__name__}.lattice_parameters should have shape "
                f"(number of spatial dimensions,)"
            )

        if len(self.atomic_basis.shape) != 2:
            raise ValueError(
                f"Invalid lattice geometry. {self.__class__.__name__}.atomic_basis should have shape "
                f"(number of particles in unit cell, number of spatial dimensions)."
            )

        if len(self.dimensions) != len(self.lattice_parameters) != self.atomic_basis.shape[1]:

            raise ValueError(
                f"Invalid lattice geometry. The number of spatial dimensions in {self.__class__.__name__}.dimensions, "
                f"{self.__class__.__name__}.lattice_parameters, and {self.__class__.__name__}.atomic_basis should "
                f"match."
            )

        if np.prod(self.lattice_parameters) <= 0:
            raise ValueError(
                "Invalid lattice geometry. The unit cell's volume, i.e. the product of lattice parameters, should be "
                "positive."
            )

    @property
    def density(self) -> float:

        r"""
        Molecular density of the solid phase, in units of molecules per volume

        Returns:
            float: density $\rho$
        """

        return self.atomic_basis.shape[0] / np.prod(self.lattice_parameters)

    @property
    def molecular_volume(self) -> float:

        r"""
        Molecular volume of the solid phase, in units of volume per molecule

        Returns:
            float: molecular volume $\omega = 1/\rho$
        """

        return 1.0 / self.density

    def initialize_simulation(self) -> Tuple[np.typing.NDArray[np.floating], np.typing.NDArray[np.floating]]:

        r"""
        Initialize the simulation by creating the lattice positions and bounds.

        Returns:
            Tuple[np.ndarray, np.ndarray]: The lattice points and supercell bounds
        """

        x = np.arange(self.dimensions[0]) * self.lattice_parameters[0]
        y = np.arange(self.dimensions[1]) * self.lattice_parameters[1]
        z = np.arange(self.dimensions[2]) * self.lattice_parameters[2]

        x_grid, y_grid, z_grid = np.meshgrid(x, y, z, indexing='ij')
        unit_cell_points = np.vstack([x_grid.ravel(), y_grid.ravel(), z_grid.ravel()]).T

        lattice_points = unit_cell_points[:, None, :] + self.atomic_basis[None, :, :] * self.lattice_parameters
        lattice_points = lattice_points.reshape(-1, 3)
        bounds = self.lattice_parameters * self.dimensions

        logger.debug("lattice sites initialized", extra={"num_sites": len(lattice_points), "bounds": bounds.tolist()})

        return lattice_points, bounds


@dataclass
class KthNearest:

    r"""
    Container class for $k$'th nearest interactions.

    This class computes the interaction matrix $\mathbf{Q}$, acting as an energy driver for the growth simulation.

    Attributes:

        cutoffs (np.ndarray):

            Interaction distance cutoffs $(\delta_1, \delta_2, \cdots, \delta_k)$. For example, two molecules with
            interaction distance between $0$ and $\delta_1$ will interact with energy $\varepsilon_1$, two molecules
            with interaction distance between $\delta_1$ and $\delta_2$ will interact with energy $\varepsilon_2$, etc.

        interaction_energies (np.ndarray):

            Interaction energies $(\varepsilon_1, \varepsilon_2, \cdots, \varepsilon_k)$.

        maxint (int):

            Optional max tree size for [KDTree](https://en.wikipedia.org/wiki/K-d_tree) distance computation. If the
            number of molecules is larger than this value, `scipy.spatial.KDTree` will switch to brute-forcing the
            distance computations.

        use_cache (bool):

            Whether to use a caching mechanism or not. If true, the simulation will try and load $\mathbf{Q}$ from
            `.kmc_cache/`. If $\mathbf{Q}$ does not exist in the local cache already, it will be stored.
    """

    cutoffs: np.typing.NDArray[np.floating]
    r"""distance cutoffs $(\delta_1, \delta_2, \cdots, \delta_k)$"""

    interaction_energies: np.typing.NDArray[np.floating]
    r"""Interaction energies $(\varepsilon_1, \varepsilon_2, \cdots, \varepsilon_k)$"""

    maxint: Optional[int] = 10_000_000
    """maximum leaf size for `scipy.spatial.KDTree`"""

    use_cache: Optional[bool] = False
    r"""Whether to use the local cache or not to fetch $\mathbf{Q}$"""

    def __post_init__(self):

        if not isinstance(self.cutoffs, np.ndarray):
            self.cutoffs = np.array(self.cutoffs)

        if not isinstance(self.interaction_energies, np.ndarray):
            self.interaction_energies = np.array(self.interaction_energies)

        if len(self.cutoffs.shape) != 1:
            raise ValueError(
                f"Invalid interaction cutoffs. {self.__class__.__name__}.cutoffs should have shape "
                f"(number of neighbors,)"
            )

        if len(self.interaction_energies.shape) != 1:
            raise ValueError(
                f"Invalid interaction energies. {self.__class__.__name__}.interaction_energies should have shape "
                f"(number of neighbors,)"
            )

        if self.cutoffs.shape != self.interaction_energies.shape:
            raise ValueError(
                f"Incompatible interaction parameters. {self.__class__.__name__}.cutoffs and "
                f"{self.__class__.__name__}.interaction_energies should have the same shape."
            )

    def compute_hamiltonian(
        self,
        lattice_points: np.typing.NDArray[np.floating],
        bounds: np.typing.NDArray[np.floating]
    ) -> scipy.sparse.csr_matrix:

        r"""
        Compute the interaction matrix $\mathbf{Q}$ from the lattice points and bounds.

        Args:
            lattice_points (np.ndarray): Lattice points in the solid phase
            bounds (np.ndarray): Supercell bounds

        Returns:
            scipy.sparse.csr_matrix: $\mathbf{Q}$ in sparse format
        """

        tree = scipy.spatial.KDTree(lattice_points, leafsize=self.maxint, boxsize=bounds)

        distance_matrix = tree.sparse_distance_matrix(tree, max_distance=self.cutoffs[-1]).tocsr()
        distance_matrix.eliminate_zeros()

        interaction_types = np.searchsorted(self.cutoffs, distance_matrix.data, side="left")
        interaction_energies = self.interaction_energies[interaction_types]

        return scipy.sparse.csr_matrix(
            (interaction_energies, distance_matrix.indices, distance_matrix.indptr),
            shape=distance_matrix.shape
        )

    def get_hamiltonian(
        self,
        lattice_points: np.typing.NDArray[np.floating],
        bounds: np.typing.NDArray[np.floating]
    ) -> scipy.sparse.csr_matrix:

        r"""
        Get the interaction matrix $\mathbf{Q}$ from the lattice points and bounds. This either calls
        `KthNearest.compute_hamiltonian`, or loads $\mathbf{Q}$ from the user's local cache.

        Args:
            lattice_points (np.ndarray): Lattice points in the solid phase
            bounds (np.ndarray): Supercell bounds

        Returns:
            scipy.sparse.csr_matrix: $\mathbf{Q}$ in sparse format
        """

        if not self.use_cache:
            hamiltonian = self.compute_hamiltonian(lattice_points, bounds)

            logger.debug("hamiltonian initialized", extra={
                "num_interactions": hamiltonian.nnz, "cohesive_energy": 0.5 * hamiltonian.sum(axis=0).mean()
            })

            return hamiltonian

        cache_folder = Path(".kmc_cache")
        cache_folder.mkdir(exist_ok=True)
        hexes = [
            array_to_hex(self.cutoffs),
            array_to_hex(self.interaction_energies),
            array_to_hex(lattice_points),
            array_to_hex(bounds)
        ]
        hamiltonian_path = cache_folder / Path(f"{'_'.join(hexes)}.npz")

        if not hamiltonian_path.exists():
            hamiltonian = self.compute_hamiltonian(lattice_points, bounds)
            scipy.sparse.save_npz(hamiltonian_path, hamiltonian)

            logger.debug("hamiltonian initialized and saved", extra={
                "num_interactions": hamiltonian.nnz, "cohesive_energy": 0.5 * hamiltonian.sum(axis=0).mean(),
                "cache_path": hamiltonian_path.name
            })

            return hamiltonian

        hamiltonian = scipy.sparse.load_npz(hamiltonian_path)

        logger.debug("hamiltonian loaded from cache", extra={
            "num_interactions": hamiltonian.nnz, "cohesive_energy": 0.5 * hamiltonian.sum(axis=0).mean(),
            "cache_path": hamiltonian_path.name
        })

        return hamiltonian
